package com.goldencode.testcases.embedded;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * Example P2J client application which establishes a secure connection to a P2J server
 * and then provides access to client spawning using a remote object.  
 */
public class CustomP2JClientApp
{
   /** Logging helper. */
   private final SimpleLogger logger;
   
   /** Internal synchronization instance. */
   private final Object lock = new Object();
   
   /** P2J server connection. */
   private Session session = null;
   
   /** Remote object used for launching client sessions. */
   private WebClientLauncher launcher = null;
   
   /**
    * Construct an instance.
    *
    * @param    logger                       
    *           Shared logging helper.  Must NOT be <code>null</code>.
    */
   public CustomP2JClientApp(SimpleLogger logger)
   {
      if (logger == null)
         throw new NullPointerException("Logger cannot be null!");
      
      this.logger = logger;
   }

   /**
    * Connect to the P2J server via SSL using the given configuration values.
    *
    * @param    hostname
    *           P2J server hostname.
    * @param    port
    *           P2J server port.
    * @param    keystore
    *           Key store filename which holds the P2J client private key used to authenticate
    *           this remote client application.
    * @param    processalias
    *           Alias in the key store which identifies the P2J client's private key.
    * @param    kspw
    *           Password used to decrypt the key store.
    * @param    kepw
    *           Password used to decrypt the private key entry.
    * @param    truststore
    *           Truststore filename which holds the P2J server certificate (allows validation of
    *           the secure connection to the P2J server).
    * @param    trustalias
    *           Alias in the truststore which identifies the P2J server's certificate.
    * @param    tspw
    *           Password used to decrypt the truststore.
    */
   public boolean connect(String hostname,
                          int    port,
                          String keystore,
                          String processalias,
                          String kspw,
                          String kepw,
                          String truststore,
                          String trustalias,
                          String tspw)
   {
      boolean result = true;
      
      try
      {
         BootstrapConfig bc = new BootstrapConfig();
         
         bc.setServer(false);
         bc.setConfigItem("net", "server", "host", hostname);
         bc.setConfigItem("net", "server", "secure_port", Integer.toString(port));
         bc.setConfigItem("net", "connection", "secure", "true");
         bc.setConfigItem("net", "dispatcher", "threads", "2");
         bc.setConfigItem("net", "queue", "conversation", "false");
         bc.setConfigItem("security", "truststore", "filename", truststore);
         bc.setConfigItem("security", "truststore", "alias", trustalias);
         bc.setConfigItem("security", "certificate", "validate", "true");
         bc.setConfigItem("security", "keystore", "filename", keystore);
         bc.setConfigItem("security", "keystore", "processalias", processalias);
         bc.setConfigItem("access", "password", "truststore", tspw);
         bc.setConfigItem("access", "password", "keystore", kspw);
         bc.setConfigItem("access", "password", "keyentry", kepw);
         
         SecurityManager secMgr = SecurityManager.createInstance(bc);
         SessionManager sessMgr = SessionManagerFactory.createLeafNode(bc);
         
         synchronized (lock)
         {
            session = sessMgr.connectDirect(null, null);
         }
         
         if (session == null)
         {
            result = false;
         }
      }
      
      catch (Exception e)
      {
         // TODO: print entire stack trace would be better
         logger.log(e.toString());
         
         e.printStackTrace();
            
         result = false;
      }
      
      return result;
   }
   
   /**
    * Disconnect the application from the P2J server. This should only be called if the
    * session is already connected.
    */
   public void disconnect()
   {
      synchronized (lock)
      {
         if (session == null)
         {
            logger.log("Cannot disconnect when not already connected!");
         }
         else
         {
            session.terminate();
         }
      }
   }
   
   /**                                       
    * Obtain a remote object instance that can be used for client launching on the P2J server.
    * This should only be called if the session is already connected.
    */
   public void importClientLauncherAPI()
   {
      synchronized (lock)
      {
         if (session == null)
         {
            logger.log("Cannot import a remote object when not already connected!");
         }
         else
         {
            launcher = (WebClientLauncher)
               RemoteObject.obtainNetworkInstance(WebClientLauncher.class, session);
         }
      }
   }
   
   /**
    * Launch a web client.  Import the remote launcher API on first usage.
    *
    * @param    user
    *           Operating system account to use.
    * @param    pw
    *           Password for the account.
    * @param    referrer
    *           When the web client exits, it will redirect back to this URL.
    *
    * @return   The URL which will securely access the spawned web client or an
    *           error message on any failure.
    */
   public String launch(String user, String pw, String referrer)
   {
      String url = null;
      
      synchronized (lock)
      {
         if (launcher == null)
         {
            importClientLauncherAPI();
         }
         
         if (launcher == null)
         {
            logger.log("Cannot obtain launcher remote object.");
         }
         else
         {
            String[] options = 
            {
               "client:cmd-line-option:param=USERID=hotel|PASSWORD=hotel",
               "client:cmd-line-option:startup-procedure=emain.p",
               "client:web:embedded=true"
            };
            url = launcher.spawn(user, pw, true, options, referrer);
         }
      }
      
      return url;
   }
}
