
package com.goldencode.testcases.embedded;

import com.goldencode.p2j.main.WebHandler;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.*;

import java.io.*;
import java.util.*;

/**
 * Command line driver to launch the example web application.
 */
public class EmbeddedWebAppDriver
{
   /** The path to the embedded web template which will be filled and returned to our client. */
   private static final String webTemplatePagePath = "/resources/index.html";
   
   /** Error web page which will be returned to our client. */
   private static final String[] failurePage =
   {
      "<!DOCTYPE html>",
      "<html>",
      "   <head>",
      "      <title>Embedded Web Client Test Harness</title>",
      "   </head>",
      "   <body style=\"background: grey\">",
      "      <div style=\"margin-top:150px;\">",
      "         <div align=\"center\" style=\"margin-top:100px\">",
      "            <h1 id=\"error\" style=\"color:red\">${url}</h1>",
      "         </div>",
      "      </div>",
      "   </body>",
      "</html>"
   };   
   
   /** P2J server hostname. */
   private String p2jHostname = "localhost";
   
   /** P2J server port. */
   private int p2jPort = -1;
   
   /** P2J client key store file name. */
   private String clientKs = null;
   
   /** P2J client key store alias that identifies the application's private key. */
   private String clientKsAlias = null;
   
   /** P2J client key store password. */
   private String clientKsPasswd = null;
   
   /** P2J client key entry password. */
   private String clientKePasswd = null;
   
   /** Trust store file name which contains the P2J server certificate. */
   private String tsFileName = null;
   
   /** Trust store password. */
   private String tsPasswd = null;
   
   /** Trust store alias that identifies the P2J server. */
   private String tsAlias = null;
   
   /** Web server port. */
   private int webPort = WebServer.DEFAULT_PORT;
   
   /** Key store file name which contains the web server's private key. */
   private String webKs = null;
   
   /** Web server key store alias that identifies the private key. */
   private String webKsAlias = null;
   
   /** Web server key store password. */
   private String webKsPasswd = null;
   
   /** Web server key entry password. */
   private String webKePasswd = null;
   
   /**
    * Parse the command line options and save the results in static variables.
    *
    * @param    args
    *           Command line arguments.
    */
   private void processOptions(String[] args)
   {
      for (int i = 0; i < args.length; i++)
      {
         // test if there is at least 1 more arg string after this one
         if ((i + 1) != args.length)
         {
            // process options that have a passed value 
            if ("-p2jhost".equals(args[i]))
            {
               p2jHostname = args[++i];
            }
            else if ("-p2jport".equals(args[i]))
            {
               p2jPort = parseInt(args[++i], -1);
               
               if (p2jPort < 1 || p2jPort > 65535)
               {
                  log(String.format("Invalid port number %d.", p2jPort));
               }
            }
            else if ("-p2jks".equals(args[i]))
            {
               clientKs = args[++i];
            }
            else if ("-p2jksalias".equals(args[i]))
            {
               clientKsAlias = args[++i];
            }
            else if ("-p2jkspw".equals(args[i]))
            {
               clientKsPasswd = args[++i];
            }
            else if ("-p2jkepw".equals(args[i]))
            {
               clientKePasswd = args[++i];
            }
            else if ("-p2jts".equals(args[i]))
            {
               tsFileName = args[++i];
            }
            else if ("-p2jtsalias".equals(args[i]))
            {
               tsAlias = args[++i];
            }
            else if ("-p2jtspw".equals(args[i]))
            {
               tsPasswd = args[++i];
            }
            else if ("-webport".equals(args[i]))
            {
               webPort = parseInt(args[++i], WebServer.DEFAULT_PORT);
               
               if (webPort < 1 || webPort > 65535)
               {
                  log(String.format("Invalid port number %d. Falling back.", webPort));
                  webPort = WebServer.DEFAULT_PORT;
               }
               else if (webPort <= 1024)
               {
                  log(String.format("WARNING: using reserved/privileged port number %d.", webPort));
               }
            }
            else if ("-webks".equals(args[i]))
            {
               webKs = args[++i];
            }
            else if ("-webksalias".equals(args[i]))
            {
               webKsAlias = args[++i];
            }
            else if ("-webkspw".equals(args[i]))
            {
               webKsPasswd = args[++i];
            }
            else if ("-webkepw".equals(args[i]))
            {
               webKePasswd = args[++i];
            }
         }
      }
   }
                           
   /**
    * Launch a web client.
    *
    * @param    user
    *           Operating system account to use.
    * @param    pw
    *           Password for the account.
    * @param    referrer
    *           When the web client exits, it will redirect back to this URL.
    *
    * @return   The URL which will securely access the spawned web client or an
    *           error message on any failure.
    */
   private String launch(String user, String pw, String referrer)
   {
      CustomP2JClientApp p2jclient = new CustomP2JClientApp(EmbeddedWebAppDriver::log);
         
      boolean success = p2jclient.connect(p2jHostname,
                                          p2jPort,
                                          clientKs,
                                          clientKsAlias,
                                          clientKsPasswd,
                                          clientKePasswd,
                                          tsFileName,
                                          tsAlias,
                                          tsPasswd);
      
      String url = null;
      
      if (success)
      {
         url = p2jclient.launch(user, pw, referrer);
         
         p2jclient.disconnect();
      }
      else
      {
         log("Failure in initialization of P2J client!");
      }
      
      return url;
   }
                                         
   /**
    * Add a message to the log. Formatting must be handled in the caller.
    *
    * @param    msg
    *           The text to add to the log.
    */
   private static void log(String msg)                          
   {
      System.out.println(msg);
   }
   
   /**
    * Helper to manage parsing strings into integers.
    *
    * @param    num                      
    *           The text to parse.
    * @param    defVal
    *           The default value to be returned on any failure.
    *
    * @return   The parsed value.
    */
   private static int parseInt(String num, int defVal)
   {
      int value = defVal;
      
      try
      {
         value = Integer.parseInt(num);
      }
      
      catch (NumberFormatException nfe)
      {
         log(String.format("NumberFormatException '%s' cannot be parsed as an integer.", num));
      }
      
      return value;
   }
   
   /**
    * Command line driver to execute the web server.
    * <p>
    * Parameters:
    * <p>
    * <pre>
    * Option          Required                Purpose                                        Default Value
    * --------------  --------  ----------------------------------------------------  -------------------------
    * -p2jhost        N         P2J Server Hostname                                   localhost
    * -p2jport        Y         P2J Server Port
    * -p2jks          Y         P2J Client Key Store Filename (contains private key)
    * -p2jksalias     Y         P2J Client Key Store Alias (identifies private key)
    * -p2jkspw        Y         P2J Client Key Store Password
    * -p2jkepw        Y         P2J Client Key Entry Password
    * -p2jts          Y         P2J Trust Store Filename (contains P2J server cert)
    * -p2jtsalias     Y         P2J Trust Store Alias
    * -p2jtspw        Y         P2J Trust Store Password
    * -webport        N         Web Server Port                                       8080
    * -webks          Y         Web Server Key Store Filename (contains private key)
    * -webksalias     Y         Web Server Key Store Alias (identifies private key)
    * -webkspw        Y         Web Server Key Store Password
    * -webkepw        Y         Web Server Key Entry Password
    * </pre>
    *
    * @param    args
    *           Command line arguments.
    */
   public static void main(String[] args)
   throws Exception
   {
      EmbeddedWebAppDriver drv = new EmbeddedWebAppDriver();
      drv.processOptions(args);
      
      log("");
      log("Configuration");
      log("---------------------------------------------");
      log(String.format("p2jhost    = %s", drv.p2jHostname));
      log(String.format("p2jport    = %d", drv.p2jPort));
      log(String.format("p2jks      = %s", drv.clientKs));
      log(String.format("p2jksalias = %s", drv.clientKsAlias));
      log(String.format("p2jkspw    = %s", drv.clientKsPasswd));
      log(String.format("p2jkepw    = %s", drv.clientKePasswd));
      log(String.format("p2jts      = %s", drv.tsFileName));
      log(String.format("p2jtsalias = %s", drv.tsAlias));
      log(String.format("p2jtspw    = %s", drv.tsPasswd));
      log(String.format("webport    = %d", drv.webPort));
      log(String.format("webks      = %s", drv.webKs));
      log(String.format("webksalias = %s", drv.webKsAlias));
      log(String.format("webkspw    = %s", drv.webKsPasswd));
      log(String.format("webkepw    = %s", drv.webKePasswd));
      log("");
      
      WebServer srv = new WebServer(EmbeddedWebAppDriver::log,
                                    drv.webPort,
                                    drv.webKs,
                                    drv.webKsPasswd,
                                    drv.webKePasswd);
      
      srv.start(new EmbeddedPageHandler(drv));
   }
   
   /**
    * Implements the Jetty handler interface which calls the P2J server to spawn a web
    * client and then it will return a page that loads that client's interface in an
    * IFRAME.
    */
   public static class EmbeddedPageHandler
   extends AbstractHandler
   {
      /** Worker to launch web clients. */
      private final EmbeddedWebAppDriver drv;
      
      /**
       * Constructor.
       */
      public EmbeddedPageHandler(EmbeddedWebAppDriver drv)
      {
         this.drv = drv;
      }
      
      /**
       * Code to process web requests. 
       */
      public void handle(String              target, 
                         Request             baseRequest,
                         HttpServletRequest  request,
                         HttpServletResponse response) 
      throws IOException,
             ServletException
      {
         System.err.println("target="+target);

         if (!(target.equals("/") || target.equals("/launch")))
         {
            baseRequest.setHandled(false);
            return;
         }
         
         response.setHeader("Access-Control-Allow-Origin", "*");
         response.setHeader("Access-Control-Allow-Methods", "POST, GET");
         response.setContentType("text/html;charset=utf-8");
         response.setStatus(HttpServletResponse.SC_OK);
         baseRequest.setHandled(true);

         if (target.equals("/launch"))
         {
            String url = drv.launch(null, null, "https://www.google.com");
            
            if (url == null || url.startsWith(WebHandler.ERROR_PREFIX))
            {
               url = "null";
            }

            response.getWriter().println(url);

            baseRequest.setHandled(true);
            return;
         }

         String[] page = loadTemplate(webTemplatePagePath);

         for (String line : page)
         {
            response.getWriter().println(line);
         }
      }
      
      /**
       * Loads the target resource template by lines.
       * 
       * @param    path
       *           The target resource path within jar file.
       * 
       * @return   The resource content by lines.
       */
      private String[] loadTemplate(String path)
      {
         LinkedList<String> lines = new LinkedList<String>();
         try
         {
            try (LineNumberReader reader = new LineNumberReader(
                     new InputStreamReader(getClass().getResourceAsStream(path))) )
            {
               String line;
               while ((line = reader.readLine()) != null)
               {
                  lines.add(line);
               }
            };
         }
         catch (IOException ex)
         {
            ex.printStackTrace();
         }
         
         return lines.toArray(new String[lines.size()]);
      }
   }
}

