package com.goldencode.testcases.embedded;

import com.goldencode.p2j.security.*;

import java.io.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.http.*;
import org.eclipse.jetty.util.resource.Resource;
import org.eclipse.jetty.util.ssl.*;

import java.util.*;
import java.security.*;
import java.security.cert.*;

import javax.net.ssl.*;

// compile with: javac -cp jetty-all-9.1.0-SNAPSHOT.jar:javax.servlet-api-3.1.0.jar:. WebServer.java
// run with: java -cp jetty-all-9.1.0-SNAPSHOT.jar:javax.servlet-api-3.1.0.jar:slf4j-jdk14-1.6.1.jar:. WebServer [options]

public class WebServer
{
   /** Default TCP/IP port to bind to (listen on). */
   public static final int DEFAULT_PORT = 8080;
   
   /** HTTP protocol in use. */
   private static final String DEFAULT_PROTOCOL = HttpVersion.HTTP_1_1.asString();
   
   /** Logging helper. */
   private final SimpleLogger logger;
   
   /** TCP/IP port for the server. */
   private int port = DEFAULT_PORT;
   
   /** Key store file name. */
   private String ksFileName = null;
   
   /** Key store password. */
   private String ksPasswd = null;
   
   /** Key entry password. */
   private String kePasswd = null;
   
   /** In-memory key store. */
   private KeyStore keyStore = null;
   
   /** Custom key manager. */
   private KeyManager[] km = null;
   
   /** Factory for configuring new SSL sessions. */
   private SslContextFactory sslContextFactory = null;
   
   /** Factory for creating new SSL connections. */
   private SslConnectionFactory sslConnectionFactory = null;
   
   /**
    * Construct an instance.
    *
    * @param    logger                       
    *           Shared logging helper.
    * @param    port
    *           The port on which to listen.
    * @param    ksFileName
    *           Key store file name.
    * @param    ksPasswd
    *           Key store password or <code>null</code> if not needed.
    * @param    kePasswd
    *           Key entry password or <code>null</code> if not needed.
    */
   public WebServer(SimpleLogger log, int port, String ksFileName, String ksPasswd, String kePasswd)
   {
      this.logger     = (log == null) ? (String msg) -> { System.err.println(msg); } : log;
      this.port       = port;
      this.ksFileName = ksFileName;
      this.ksPasswd   = ksPasswd;
      this.kePasswd   = kePasswd;
      
      if (port < 1 || port > 65535)
      {
         logger.log(String.format("Invalid port number %d. Falling back.", port));
         port = DEFAULT_PORT;
      }
      else if (port <= 1024)
      {
         logger.log(String.format("WARNING: using reserved/privileged port number %d.", port));
      }
   }
   
   /**
    * Start the web server and wait for it to end.  The calling thread will block until the
    * web server exits!
    *
    * @param    handler
    *           The worker instance that will process all web requests.
    */
   public void start(Handler handler)
   throws Exception
   {
      Server server = new Server();
      
      if (!initSSL())
      {
         return;
      }
      
      setupSSLConnector(server);
      
      HandlerCollection list = new HandlerCollection();
      
      ResourceHandler commonRes = new ResourceHandler();
      commonRes.setDirectoriesListed(false);
      commonRes.setResourceBase(".");
      
      list.setHandlers(new Handler[] { handler, commonRes, new DefaultHandler() });
      
      server.setHandler(list);
      
      server.start();
      server.join();
   }
   
   /**
    * Initialize SSL by loading key material.
    *
    * @return   <code>true</code> if everything was initialized successfully.
    */
   private boolean initSSL()
   {
      try
      {
         keyStore = KeyStore.getInstance("JKS");
      }
      
      catch (KeyStoreException kse)
      {
         logger.log(String.format("JKS KeyStore cannot be initialized (%s).", kse.toString()));
         return false;
      }
      
      SSLHelper.ErrorCode rc = SSLHelper.loadKeyStoreFromFile(keyStore, ksFileName, ksPasswd);
      
      if (rc != SSLHelper.ErrorCode.NO_ERROR)
      {
         logger.log(String.format("KeyStore load failure (%s).", rc.toString()));
         return false;
      }
      
      Object[] result = new Object[1];
      
      rc = SSLHelper.initKeyManagers(keyStore, kePasswd, result);
      
      if (rc != SSLHelper.ErrorCode.NO_ERROR)
      {
         logger.log(String.format("KeyManager[] init failure (%s).", rc.toString()));
         return false;
      }
      
      km = (KeyManager[]) result[0];
      
      SSLContext ctx = null;
      
      try
      {
         ctx = SSLContext.getInstance("TLS");
         ctx.init(km, null, null);
      }
      
      catch (NoSuchAlgorithmException nsa)
      {
         logger.log(String.format("SSLContext 'TLS' cannot be instantiated (%s).", nsa.toString()));
         return false;
      }
      catch (KeyManagementException kme)
      {
         logger.log(String.format("SSLContext cannot be initialized (%s).", kme.toString()));
         return false;
      }
      
      sslContextFactory = new SslContextFactory();
      sslContextFactory.setSslContext(ctx);
      
      sslConnectionFactory = new SslConnectionFactory(sslContextFactory, DEFAULT_PROTOCOL);
      
      return true;
   }
   
   /**
    * Initialize the SSL connector and attach it to the given server instance. This will
    * default the host to all interfaces.
    *
    * @param    server
    *           Web server instance.
    */
   private void setupSSLConnector(Server server)
   {
      HttpConnectionFactory hcf = new HttpConnectionFactory(new HttpConfiguration());
      ServerConnector       sc  = new ServerConnector(server, sslConnectionFactory, hcf);
      
      // TODO: what happens if we don't have the privledges to open this port?
      sc.setPort(port);
      server.addConnector(sc);
   }
}
