/*
 * Copyright 2016-2017 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * This function is loaded by the Hotel embedded web application's index.html. It defines the
 * primary JavaScript API for the application.
 */
define(function ()
{
   "use strict";
   
   /** Object which contains API functions. */
   var me = {};
   
   /** The generator for the message IDs, on the remote side (always positive). */ 
   var nextMessageId = 0;
   
   /**
    * The name of the program where the APIs are located.  This will be ran persistent once the
    * P2J iframe has loaded.
    */
   var remoteProgram = "ehotel.p";
   
   /** The handle for the remote program, in which the APIs are executed. */
   var remoteProgramHandle;
   
   /** Map of parsers for each API. */
   var incomingRequestParser = {};
   
   /** Map of handlers for each API. */
   var incomingRequestHandlers = {};
   
   /** Current start date for room utilization by type data. */
   var utilByTypeStartDate;
   
   /** Current end date for room utilization by type data. */
   var utilByTypeEndDate;
   
   /** Debugging flag. */
   var DEBUG = false;
   
   /** User name. */
   var fwd_u = "";
   
   /** User password. */
   var fwd_p = "";
   
   /** Should application be reloaded after user disconnection from FWD client driver? */
   var refreshOnDisconnect = true;
   
   /** Resource ID for a 'roomUtilization' pub/sub event published by converted ABL code. */
   var roomUtilizationResourceId;
   
   /** Resource ID for a 'capacityZoom' pub/sub event published by JavaScript code. */
   var capacityZoomResourceId;
   
   /** Resource ID for a 'roomTypePick' pub/sub event published by JavaScript code. */
   var roomTypePickResourceId;
   
   // TODO: move this up-call machinery into FWD runtime
   
   // map up-call response types to JSON message parsers that know how to interpret the response
   incomingRequestParser["executeCommandResponse"]    = parseExecuteCommandResponse;
   incomingRequestParser["utilizationByDateResponse"] = parseUtilizationByDateResponse;
   incomingRequestParser["utilizationByTypeResponse"] = parseUtilizationByTypeResponse;
   incomingRequestParser["p2jInitialized"]            = parseInvokeRemoteProgram;
   incomingRequestParser["showError"]                 = parseShowError;
   
   // map up-call response types to callback functions that handle the response
   incomingRequestHandlers["executeCommandResponse"]    = executeCommandResponse;
   incomingRequestHandlers["utilizationByDateResponse"] = utilizationByDateResponse;
   incomingRequestHandlers["utilizationByTypeResponse"] = utilizationByTypeResponse;
   incomingRequestHandlers["p2jInitialized"]            = invokeRemoteProgram;
   incomingRequestHandlers["showError"]                 = showError;
   
   /**
    * Format the given date into the format expected by ABL (or return today if no date is passed).
    *
    * @param   {Date} date
    *          The date to format.
    */
   me.formattedDate = function(date)
   {
      if (!date)
      {
         date = new Date(Date.now());
      }
      
      return moment(date).format("MM/DD/YYYY");
   }
   
   /**
    * Parse a Progress date string (in the format 99/99/999) and return the corresponding date.
    * 
    * TODO: make this honor the current SESSION:DATE-FORMAT setting (mdy/dmy); currently, it
    * assumes mdy.
    */
   me.parseProgressDate = function(dateString)
   {
      // assume Progress date format of 99/99/9999 (mdy)
      var a = dateString.split(/[^0-9]/);
      
      return new Date(a[2], a[0] - 1, a[1]);
   }
   
   /**
    * Enable/Disable the debug of incoming messages.
    *
    * @param   {boolean} state
    *          The debug state.
    */
   me.setDebug = function(state)
   {
      DEBUG = state;
   }
   
   /**
    * Callback executed when the P2J page has loaded.  It runs the remote program persistent.
    */
   me.initialize = function()
   {
      // do not call the remote program here, wait for the "p2jStarted" notification
      p2j.embedded.silentErrorEnable();

      // examples for UNSUBSCRIBE and PUBLISH:
      // p2j.embedded.unsubscribe(unsubscribeCallback, true, loggedinResourceEvent, "loggedin", undefined);
      // p2j.embedded.publish("loggedin", loggedinResourceEvent, ["hotel2", "hotel2"]);
      
      // subscribe to a "roomUtilization" event published by converted ABL code
      p2j.embedded.subscribe(subscribeRoomUtilizationCallback,
                             true,
                             roomUtilizationResourceId,
                             "roomUtilization",
                             undefined,
                             roomUtilizationEvent);
      
      // create a resource for a "capacityZoom" event that will be published from JavaScript
      p2j.embedded.createResource(capacityZoomResourceCreated);
      
      // create a resource for a "roomTypePick" event that will be published from JavaScript
      p2j.embedded.createResource(roomTypePickResourceCreated);
      
      // perform login
      var msgId = getNextMessageId();
      var args = 
      {
         "messageId": msgId.toString(),
         "username" : fwd_u,
         "password" : fwd_p
      };
      p2j.embedded.invokeProcedure(null, "doLogin", args, null);
   };
   
   /**
    * Callback executed for an UNSUBSCRIBE response.
    *
    * Not currently in use; provided as an example.
    *
    * @param   {object} msg
    *          The response message.
    */
   function unsubscribeCallback(msg)
   {
      var payload = msg["payload"];
      if (payload["error"])
      {
         console.log(payload["error-num"] + " " + payload["error-msg"]);
      }
      showMessage("unsubscribeCallback", msg);
   };
   
   /**
    * Callback executed for a 'roomUtilization' subscribe response.
    *
    * @param   {object} msg
    *          The response message.
    */
   function subscribeRoomUtilizationCallback(msg)
   {
      var payload = msg["payload"];
      if (payload["error"])
      {
         console.log(payload["error-num"] + " " + payload["error-msg"]);
      }

      showMessage("subscribeRoomUtilizationCallback", msg);

      if (roomUtilizationResourceId)
      {
         return;
      }
      
      roomUtilizationResourceId = msg["payload"]["subscriber-id"];
   }
   
   /**
    * Callback executed for creation of a resource associated with the 'capacityZoom' pub/sub
    * event.
    *
    * @param   {object} msg
    *          The response message.
    */
   function capacityZoomResourceCreated(msg)
   {
      showMessage("capacityZoomResourceCreated", msg);
      var payload = msg["payload"];
      capacityZoomResourceId = msg["payload"]["resource-id"];
   }
   
   /**
    * Callback executed for creation of a resource associated with the 'roomTypePick' pub/sub
    * event.
    *
    * @param   {object} msg
    *          The response message.
    */
   function roomTypePickResourceCreated(msg)
   {
      showMessage("roomTypePickResourceCreated", msg);
      var payload = msg["payload"];
      roomTypePickResourceId = msg["payload"]["resource-id"];
   }
   
   /**
    * Callback executed for the "roomUtilization" PUBLISH event. This event is published by
    * converted ABL code when either a new hotel check-in or a new hotel reservation is booked.
    *
    * @param   {array} params
    *          The publish event parameters.
    */
   function roomUtilizationEvent(params)
   {
      // start date is check-in date of stay/reservation
      var start = ehotel.parseProgressDate(params[0]);
      
      // end date is day before check-out date of stay/reservation
      var end = moment(ehotel.parseProgressDate(params[1])).subtract(1, "days").toDate();
      
      // up-call to collect utilization/capacity data for given date window
      ehotel.getUtilizationByDate(ehotel.formattedDate(start), ehotel.formattedDate(end));
      
      // up-call to collect utilization data by room type for given date window; only invoke
      // if the currently displayed date window overlaps with the period that changed
      if (utilByTypeStartDate <= end && utilByTypeEndDate >= start)
      {
         ehotel.getUtilizationByType(utilByTypeStartDate, utilByTypeEndDate);
      }
   }
   
   /**
    * Generic handler for the messages incoming from the converted code.  These can be either
    * responses for a INVOKE request or REMOTE_CALL requests.
    *
    * @param   {object} msg
    *          The received message.
    */
   me.messageCallback = function(msg)
   {
      console.log(msg);
      
      if (msg.type == p2j.embedded.REMOTE_CALL)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            // do the work
            handleIncomingRequest(msg.payload);
            
            // send just an OK response
            p2j.embedded.sendRemoteCallResponse(msg.id, "OK");
         }
         else
         {
            // a RESPONSE mode for REMOTE_CALL should never happen 
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a REMOTE_CALL message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
      }
      else if (msg.type == p2j.embedded.INVOKE)
      {
         if (msg.mode == p2j.embedded.RESPONSE)
         {
            if (msg.payload.error)
            {
               console.error(msg.payload["error-num"] + " " + msg.payload["error-msg"]);
            }
            
            // otherwise, don't do anything
         }
         else
         {
            // a REQUEST mode for INVOKE should never happen
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a INVOKE message with " + msgMode + 
                        " mode - remote side can handle only RESPONSE modes!");
         }
      }
   };

   /**
    * Callback executed when the P2J (FWD) page has closed.
    *
    * @param    {string} origin
    *           The origin of the P2J page.
    */
   me.pageClosed = function(origin)
   {
      dbg("page closed: " + origin);
      setTestState(false);
   };
   
   /**
    * Callback executed when the P2J (FWD) page has loaded.
    *
    * @param    {string} origin
    *           The origin of the P2J page.
    */
   me.pageLoaded = function(origin)
   {
      ehotel.initialize();
   };
   
   /**
    * Callback executed when the P2J (FWD) socket has disconnected.
    */
   me.socketDisconnected = function()
   {
      dbg("socket disconnected");
         
      setTestState(false);
   };
  
   /**
    * Callback executed when the P2J (FWD) socket has reconnected.
    */
   me.socketReconnected = function()
   {
      dbg("socket reconnected");
      
      setTestState(true);
   };

   /**
    * Notify the remote side that the ehotel session is ending.
    */
   me.notifyEHotelDisconnect = function()
   {
      if (remoteProgramHandle)
      {
         // no arguments are passed to this API
         p2j.embedded.invokeProcedure(refreshPage, "inMsg_notifyEHotelDisconnect", {}, remoteProgramHandle);
      }
      else
      {
         // no arguments are passed to this API
         p2j.embedded.invokeProcedure(refreshPage, "inMsg_notifyEHotelDisconnect", {}, null);
      }
      
      $(".nav > li a", "#topNavBar").hide();
      
      var mainView = $("#mainView");
      mainView.css("opacity", 0);
      setTimeout(function ()
      {
         mainView.hide();
      }, 1000);

      $("#bgFader").css("opacity", 0);
   };
   
   /**
    * Execute the "executeCommand" API.
    *
    * @param   {String} command
    *          The command to be executed.
    */
   me.executeCommand = function(command)
   {
      var args = 
      {
         "command" : command 
      };
      
      invoke("executeCommand", args);
   };
   
   /**
    * Send a command to load the room availability table.
    *
    * @param   {String} adate
    *          The date on which the availability is computed.
    */
   me.loadRooms = function(adate)
   {
      var args =
      {
         "adate" : adate
      }
      
      invoke("loadRooms", args);
   }
   
   /**
    * Make an up-call to converted ABL code to collect room utilization data, grouped by room
    * type, aggregated for the specified date range.
    *
    * @param   {date} start
    *          Starting date.
    * @param   {date} end
    *          Ending date.
    */
   me.getUtilizationByType = function(start, end)
   {
      var args = 
      {
         "start-date" : ehotel.formattedDate(start),
         "end-date"   : ehotel.formattedDate(end)
      };

      invoke("utilizationByType", args);
   };
   
   /**
    * Make an up-call to converted ABL code to collect room utilization data for each date in
    * the specified range.
    *
    * @param   {string} start
    *          A Progress-formatted date string for the start of the range.
    * @param   {string} end
    *          A Progress-formatted date string for the end of the range.
    */
   me.getUtilizationByDate = function(start, end)
   {
      var args = 
      {
         "start-date" : start,
         "end-date"   : end
      };

      invoke("utilizationByDate", args);
   };
   
   /**
    * This function is invoked from within the calendar heat map when the user clicks a date or
    * swipes a date range in the calendar control. It updates the aster plot, room utilization
    * table, and available rooms embedded content view (if showing) to display information
    * specific to the selected date range.
    *
    * This causes a 'capacityZoom' event to be published.
    *
    * @param   {date} start
    *          Starting date for the selected range; will always be less than or equal to end
    *          date, regardless of the direction the user swiped.
    * @param   {date} end
    *          Ending date for the selected range; will always be greater than or equal to start
    *          date, regardless of the direction the user swiped.
    *          
    */
   me.dateRangeSelected = function dateRangeSelected(start, end)
   {
      // refresh aster plot and room utilization table of room utilization by type
      ehotel.getUtilizationByType(start, end);
      
      // notify subscribers that a date range on the hotel capacity heatmap has been selected
      var a = ehotel.formattedDate(start);
      var b = ehotel.formattedDate(end);
      p2j.embedded.publish("capacityZoom", capacityZoomResourceId, [a, b]);
   }
   
   /**
    * Log into the client driver in order to load the embedded, converted ABL content.
    *
    * @param   {string} u
    *          User name.
    * @param   {string} p
    *          Password.
    */
   me.login = function(u, p)
   {
      var ifr = document.getElementById("embeddedP2J");

      fwd_u = u;
      fwd_p = p;
      
      if (ifr.src.startsWith("https://"))
      {
         var msgId = getNextMessageId();

         var args = 
         {
            "messageId": msgId.toString(),
            "username" : fwd_u,
            "password" : fwd_p
         };

         p2j.embedded.invokeProcedure(null, "doLogin", args, null);
      }
      else
      {
         var xmlHttp = new XMLHttpRequest();
         xmlHttp.onreadystatechange = function()
         { 
            if (xmlHttp.readyState == 4 && xmlHttp.status == 200)
            {
               var fwd_url = xmlHttp.responseText;
               
               if (fwd_url.startsWith("null"))
               {
                  showError({ "error": "Could not connect to the FWD server!"} );
               }
               else
               {
                  ifr.src = fwd_url;
                  
                  $("#p2jLoginContainer").hide();
   
                  // the main view is transparent at this point; however, we need to realize it here to
                  // work around a firefox bug (https://bugzilla.mozilla.org/show_bug.cgi?id=941146),
                  // which we would otherwise hit due to FWD's use of an unrealized canvas to calculate
                  // some font metrics
                  $("#mainView").show();
                  
                  // mostly hide background image with a nearly opaque overlay
                  $("#bgFader").show().css("opacity", 0.87);
                  
                  p2j.embedded.init("embeddedP2J", "embeddedP2JOverlay", "embedded-overlay");
                  p2j.embedded.setMessageCallback(ehotel.messageCallback);
                  p2j.embedded.setPageLoadedCallback(ehotel.pageLoaded);
                  p2j.embedded.setPageClosedCallback(ehotel.pageClosed);
                  p2j.embedded.setSocketDisconnectedCallback(ehotel.socketDisconnected);
                  p2j.embedded.setSocketReconnectedCallback(ehotel.socketReconnected);
               }
            }
         }
         
         xmlHttp.open("GET", "/launch", true); 
         xmlHttp.send(null);
      }
   };
   
   /**
    * Execute the specified API, with the specified arguments.  This will execute the 
    * p2j.embedded.invokeProcedure API, with the program name set to "inMsg_" + apiName.  Also, 
    * a "messageId" is added to the arguments.
    *
    * @param    {string} apiName
    *           The name of the remote API to execute.
    * @param    {object} args
    *           The argument map.  The values must be strings.
    */
   function invoke(apiName, args)
   {
      if (remoteProgramHandle)
      {
         var msgId = getNextMessageId();
         args["messageId"] = msgId.toString();
   
         p2j.embedded.invokeProcedure(null, "inMsg_" + apiName, args, remoteProgramHandle);
      }
   };
   
   /**
    * After the remote program (i.e., the embedded, converted ABL program) has been invoked,
    * this callback is invoked. Store the remote program handle, and perform any required
    * initialization of the embedding web application here.
    *
    * @param   {object}
    *          The message response.
    */
   function saveRemoteProgramHandle(msg)
   {
      if (msg.payload.error)
      {
         dbg(msg.payload["error-num"] + " " + msg.payload["error-msg"]);
      }

      remoteProgramHandle = msg.payload.handle;
      
      setTestState(true);
      
      // get the initial set of room by type data for the current date, to populate the aster
      // plot and tabulator table
      var d = new Date();
      ehotel.getUtilizationByType(d, d);
      
      // activate the first menu item by default
      activateMenu("#test1");
      ehotel.executeCommand("showAvailRooms");
      $('#topNavMenu').collapse("hide");
      
      // show navbar menu items
      $(".nav > li a", "#topNavBar").show();
      
      $("#mainView").css("opacity", 1);
      $("#waitMsg").css("opacity", 0).text("Loading...");
      
      // initialize the tabulator
      $("#utilTypeTable").tabulator(
      {
         fitColumns: true, // fit columns to width of table
         height: "100%",
         columns:
         [
            {
               title:"Type",
               field:"label",
               tooltip:true,
               tooltipHeader:"Room Type",
               sorter:"string",
            },
            {
               title:"Room Nights",
               columns:
               [
                  {
                     title:"Actual",
                     field:"roomsOccupied",
                     tooltipHeader:"Rented Room Nights",
                     sorter:"number",
                     align:"right",
                     width:75,
                  },
                  {
                     title:"Total",
                     field:"totalRooms",
                     tooltipHeader:"Total Room Nights",
                     sorter:"number",
                     align:"right",
                     width:75,
                  },
               ],
            },
            {
               title:"Revenue",
               columns:
               [
                  {
                     title:"Actual",
                     field:"actualRevenue",
                     tooltip:true,
                     tooltipHeader:"Actual Revenue",
                     sorter:"number",
                     align:"right",
                     width:90,
                  },
                  {
                     title:"Poten.",
                     field:"potentialRevenue",
                     tooltip:true,
                     tooltipHeader:"Potential Total Revenue",
                     sorter:"number",
                     align:"right",
                     width:90,
                  },
               ],
            },
            {
               title:"%",
               field:"score",
               tooltip:true,
               tooltipHeader:"Room Type Utilization",
               sorter:"number",
               formatter:"progress",
               width:50,
            },
         ],
         rowClick:function(e, id, data, row)
         {
            // notify subscribers that a room type has been picked in table
            var i = id.toString();
            var a = ehotel.formattedDate(utilByTypeStartDate);
            var b = ehotel.formattedDate(utilByTypeEndDate);
            p2j.embedded.publish("roomTypePick", roomTypePickResourceId, [i, a, b]);
         },
      });
      
      // re-render table on window resize
      $(window).resize(function()
      {
         $("#utilTypeTable").tabulator("redraw", true);
      });

      // get date range calendar can render within its parent's width
      var calDates = ehotel.calendar.initialize(new Date(), $("#capacityCalParent").width());
      
      // query daily utilization data
      ehotel.getUtilizationByDate(ehotel.formattedDate(calDates[0]),
                                  ehotel.formattedDate(calDates[1]));
   };
   
   /** 
    * Generic handler for incoming REMOTE_CALL requests.  It delegates the call to a dedicated
    * handler, based on the "api" value.
    * 
    * TODO: move this processing into the FWD runtime.
    * 
    * @param   {string} msg
    *          The request message.
    */
   function handleIncomingRequest(msg)
   {
      var msgJson = JSON.parse(msg);
      
      var api = msgJson["api"];

      if (incomingRequestParser[api] == undefined)
      {
         dbg("Received message " + msg + " with unknown handler for " + api);
         return;
      }

      // parse the message
      msgJson = incomingRequestParser[api](msg);
      
      // invoke the appropriate handle
      incomingRequestHandlers[api](msgJson);
   };
   
   /**
    * Parser for the "sendMsg_executeCommandResponse" REMOTE_CALL request.
    *
    * TODO: move this processing into the FWD runtime.
    * 
    * @param   {string} msg
    *          The incoming message.
    * @return  {object}
    *          The JSON object.
    */
   function parseExecuteCommandResponse(msg)
   {
      var jsonMsg = JSON.parse(msg);
      
      var api = jsonMsg.api;
      var cmd = jsonMsg.command;
      var messageId = jsonMsg.messageId;
      
      var omsg =
      {
         "messageId" : parseInt(messageId),
         "command" : cmd
      };
      
      return omsg;
   };
   
   /**
    * Parser for the "sendMsg_utilizationByDateResponse" REMOTE_CALL request.
    *
    * @param   {string} msg
    *          The incoming message.
    *
    * @return  {object}
    *          The JSON object.
    */
   function parseUtilizationByDateResponse(msg)
   {
      var jsonMsg = JSON.parse(msg);
      var messageId = jsonMsg.messageId;
      var omsg =
      {
         "messageId"  : parseInt(messageId),
         "rows"       : jsonMsg.rows,
         "domain"     : jsonMsg.domain,
         "startDate"  : jsonMsg.start,
         "endDate"    : jsonMsg.end
      };
      
      return omsg;
   }
   
   /**
    * Parser for the "sendMsg_utilizationByTypeResponse" REMOTE_CALL request.
    *
    * @param   {string} msg
    *          The incoming message.
    *
    * @return  {object}
    *          The JSON object.
    */
   function parseUtilizationByTypeResponse(msg)
   {
      var jsonMsg = JSON.parse(msg);
      var messageId = jsonMsg.messageId;
      var omsg =
      {
         "messageId"    : parseInt(messageId),
         "rows"         : jsonMsg.rows,
         "rateWeighted" : (jsonMsg.rateWeighted == "true"),
         "start"        : jsonMsg.start,
         "end"          : jsonMsg.end
      };
      
      return omsg;
   }
   
   /**
    * Parser for the "showError" REMOTE_CALL request.
    *
    * @param   {string} msg
    *          The incoming message.
    *
    * @return  {object}
    *          The JSON object.
    */
   function parseShowError(msg)
   {
      var jsonMsg = JSON.parse(msg);

      return jsonMsg;
   };

   /**
    * Parser for the "p2jInitialized" REMOTE_CALL notification.
    *
    * @param   {string} msg
    *          The incoming message.
    * @return  {object}
    *          The JSON from the parsed message.
    */
   function parseInvokeRemoteProgram(msg)
   {
      // nothing to do, this is just a notification
      return {};
   };
   
   /**
    * Handler for the "sendMsg_executeCommandResponse" REMOTE_CALL request.
    *
    * TODO: move this processing into the FWD runtime.
    * 
    * @param   {object} msg
    *          The JSON from the parsed message.
    */
   function executeCommandResponse(jsonMsg)
   {
      var messageId = jsonMsg.messageId;
      var cmd = jsonMsg.command;

      showMessage("executeCommandResponse", jsonMsg);
   };
   
   /**
    * For the "p2jInitialized" REMOTE_CALL notification, it invokes the remote program.
    *
    * @param   {object} msg
    *          An empty object.
    */
   function invokeRemoteProgram(msg)
   {
      p2j.embedded.invokeExternalProgramPersistent(saveRemoteProgramHandle, remoteProgram, {});
   };
   
   /**
    * Generate a new message ID for a request into the P2J client.
    *
    * @return   a message ID: always a positive value.
    */
   function getNextMessageId()
   {
      // the messages incoming from P2J have a negative value; the messages from other windows
      // must have a positive value
      var msgId = nextMessageId + 1;
      
      if (msgId < 0)
      {
         msgId = 1;
      }
      
      nextMessageId = msgId;
      
      return msgId;
   };

   /**
    * Show the message in the specified JSON in an alert box.
    *
    * @param   {string} apiName
    *          The API to which the message belongs.
    * @param   {object} msg
    *          The JSON object with the message.
    */
   function showMessage(apiName, msg)
   {
      var smsg = JSON.stringify(msg);
      dbg(apiName + " " + smsg);
   }
   
   /**
    * Express a message, either as an alert (in debug mode), or as an entry in the console log.
    */
   function dbg(msg)
   {
      if (DEBUG)
      {
         alert(msg);
      }
      else
      {
         console.error(msg);
      }
   }

   /**
    * Enable/disable the test buttons.
    *
    * @param   {boolean} state
    *          Flag indicating if the buttons should be enabled or disabled.
    */
   function setTestState(state)
   {
      document.getElementById("test1").disabled = !state;
      document.getElementById("test2").disabled = !state;
      document.getElementById("test3").disabled = !state;
      document.getElementById("test4").disabled = !state;
      document.getElementById("test5").disabled = !state;
      document.getElementById("testX").disabled = !state;
   }
   
   /**
    * Reload the current page after a slight delay.
    */
   function refreshPage()
   {
      setTimeout(function ()
      {
         location.reload(true);
      }, 750);
   };
   
   /**
    * Set the URL for the iframe containing embedded, converted ABL content to 'about:none'.
    */
   function setIframeToNone()
   {
      p2j.embedded.handleMessages = false;
      var ifr = document.getElementById("embeddedP2J");
      ifr.src = "about:none";
      
      refreshOnDisconnect = true;
   }
   
   /**
    * Show an error from a JSON message in the alert area of the login container.
    *
    * @param   {object} jsonMsg
    *          Object containing error message.
    */
   function showError(jsonMsg)
   {
      if (jsonMsg.error == "Invalid username or password")
      {
         $("#bgFader").css("opacity", 0);
         $("#p2jLoginContainer").show();
         $("#mainView").hide();
         refreshOnDisconnect = false;
         p2j.embedded.invokeProcedure(setIframeToNone, "inMsg_notifyEHotelDisconnect", {}, null);
      }

      $("#p2jAlertArea").text(jsonMsg.error).show();
   }
   
   /**
    * Handler for the "sendMsg_utilizationByDateResponse" REMOTE_CALL request.
    *
    * Updates "heat map" calendar with utilization data for each day displayed by calendar.
    *
    * @param   {object} msg
    *          The JSON from the parsed message.
    */
   function utilizationByDateResponse(jsonMsg)
   {
      // this will create the calendar control if needed, or update data if already present
      ehotel.calendar.render(jsonMsg,
                             document.getElementById("capacityCalendar"),
                             ehotel.dateRangeSelected,
                             "EMPTY",
                             "FULL");
   }

   /**
    * Handler for the "sendMsg_utilizationByTypeResponse" REMOTE_CALL request.
    *
    * Draws an aster plot of utilization by room type. Populates utilization table with data.
    *
    * @param   {object} msg
    *          The JSON from the parsed message.
    */
   function utilizationByTypeResponse(jsonMsg)
   {
      // remove any previously generated plot
      $("#asterPlot").empty();
      
      utilByTypeStartDate = ehotel.parseProgressDate(jsonMsg.start);
      utilByTypeEndDate = ehotel.parseProgressDate(jsonMsg.end);
      
      var fmt = "ddd MMM DD YYYY";
      var txt = moment(utilByTypeStartDate).format(fmt);
      if (utilByTypeStartDate < utilByTypeEndDate)
      {
         txt = txt + " - " + moment(utilByTypeEndDate).format(fmt);
      }
      $("#utilTypePeriod").text(txt);
      
      var rows = jsonMsg.rows;
      
      /* TODO: kludgy, will fail if > 14 room types */
      var colors =
      [
         "#9E0041",
         "#C32F4B",
         "#E1514B",
         "#F47245",
         "#FB9F59",
         "#FEC574",
         "#FAE38C",
         "#EAF195",
         "#C7E89E",
         "#9CD6A4",
         "#6CC4A4",
         "#4D9DB4",
         "#4776B4",
         "#5E4EA1"
      ];
      
      // assign arc colors
      for (var i = 0; i < rows.length; i++)
      {
         rows[i].color = colors[i];
      }
      
      // determine which title to use
      if (jsonMsg.rateWeighted == true)
      {
         $("#utilRoomTypeLabel").text("MEDIAN REVENUE EARNED BY ROOM TYPE");
      }
      else
      {
         $("#utilRoomTypeLabel").text("MEDIAN OCCUPANCY BY ROOM TYPE");
      }
      
      // generate the new plot
      ehotel.aster.generate("#asterPlot", rows);
      
      // set data into utilization table
      $("#utilTypeTable").tabulator("setData", rows);
   }
   
   return me;
});
